@font-face {
  font-family: 'Lato';
  src:
    url('./lato-v14-latin/lato-v14-latin-regular.woff2'),
    url('./lato-v14-latin/lato-v14-latin-regular.woff'),
    url('./lato-v14-latin/lato-v14-latin-regular.ttf'),
    url('./lato-v14-latin/lato-v14-latin-regular.svg');
  font-style: 'normal';
  font-weight: 400;
}

@font-face {
  font-family: 'Lato';
  src:
    url('./lato-v14-latin/lato-v14-latin-700.woff2'),
    url('./lato-v14-latin/lato-v14-latin-700.woff'),
    url('./lato-v14-latin/lato-v14-latin-700.ttf'),
    url('./lato-v14-latin/lato-v14-latin-700.svg');
  font-style: 'normal';
  font-weight: 700;
}

@font-face {
  font-family: 'IBM Plex Mono';
  src:
    url('./ibm-plex-mono-v4-latin/ibm-plex-mono-v4-latin-regular.woff2'),
    url('./ibm-plex-mono-v4-latin/ibm-plex-mono-v4-latin-regular.woff'),
    url('./ibm-plex-mono-v4-latin/ibm-plex-mono-v4-latin-regular.ttf'),
    url('./ibm-plex-mono-v4-latin/ibm-plex-mono-v4-latin-regular.svg');
  font-style: 'normal';
  font-weight: 400;
}
