DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256

LOG_COMPONENT_SOURCE_METHOD_INIT = "init"
LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON = "fingerprint_addon"
