import{p as e,f as o}from"./flowDb-956e92f1-67235ff6.js";import{f as t,g as a}from"./styles-c10674c1-8e90dbb9.js";import{u as i}from"./index-051c5a6e.js";import"./graph-f3c1d212.js";import"./layout-e93126bc.js";import"./index-3862675e-34cbca30.js";import"./clone-281a0990.js";import"./edges-e0da2a9e-7b25b4af.js";import"./createText-2e5e7dd3-3751dffe.js";import"./line-15eb1e26.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-d2444dfd.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
